
from __future__ import absolute_import
import copy

from ansible.utils import check_conditional, plugins
from ansible import errors

from ansibledebugger.constants import WRAPPED_ACTION_PLUGIN_SUFFIX
from ansibledebugger.interpreter import TaskInfo, ErrorInfo, NextAction
from ansibledebugger.interpreter.simple_interpreter import Interpreter


class ActionModule(object):
    """ An action plugin wrapper which invokes a debugger when a task fails """
    def __init__(self, runner):
        plugin_name = self.__module__.split('.')[-1]
        wrapped_plugin_name = plugin_name + WRAPPED_ACTION_PLUGIN_SUFFIX
        wrapped_plugin = plugins.action_loader.get(wrapped_plugin_name, runner)

        self.__class__ = type(wrapped_plugin.__class__.__name__,
                              (self.__class__, wrapped_plugin.__class__), {})
        self.__dict__ = wrapped_plugin.__dict__

        self.wrapped_plugin = wrapped_plugin
        self.runner = runner

    def run(self, conn, tmp_path, module_name, module_args, inject, complex_args=None, **kwargs):
        task_info = TaskInfo(conn, tmp_path, module_name, module_args, inject, complex_args)
        return_data, error_info = self._run(task_info, **kwargs)

        while error_info.failed:
            next_action = self._show_interpreter(task_info, return_data, error_info)
            if next_action.result == NextAction.REDO:
                return_data, error_info = self._run(task_info, **kwargs)

            elif next_action.result == NextAction.CONTINUE or next_action.result == NextAction.EXIT:
                # CONTINUE and EXIT are same so far
                if error_info.error is not None:
                    raise error_info.error
                else:
                    break

        return return_data

    def _run(self, task_info, **kwargs_for_run):
        try:
            return_data = self.wrapped_plugin.run(task_info.conn, task_info.tmp_path, task_info.module_name,
                                                  task_info.module_args, task_info.vars,
                                                  task_info.complex_args, **kwargs_for_run)

            ignore_errors = task_info.vars.get('ignore_errors', False)
            failed_when = task_info.vars.get('failed_when', None)
            error_info = self._is_failed(return_data, ignore_errors, failed_when, task_info)

        except errors.AnsibleError, ae:
            return_data = None
            error_info = ErrorInfo(True, errors.AnsibleError.__name__, str(ae), ae)

        return return_data, error_info

    def _is_failed(self, return_data, ignore_errors, failed_when_cond, task_info):
        """Check the result of task execution. """
        failed = False
        reason = None

        result = return_data.result
        failed_flag = result.get('failed', False)
        error_rc = (result.get('rc', 0) != 0)

        if failed_when_cond is not None:
            failed_when_result, reason = self._check_failed_when(failed_when_cond, task_info, result)
        else:
            failed_when_result = False

        if not ignore_errors:
            if failed_flag:
                failed = True
                reason = 'the task returned with a "failed" flag'
            else:
                if failed_when_cond is not None:
                    if failed_when_result:
                        failed = True
                        # reason is already set
                else:
                    if error_rc:
                        failed = True
                        reason = 'return code is not 0'

        return ErrorInfo(failed, reason, str(return_data.result))

    def _check_failed_when(self, failed_when, task_info, return_data):
        """Check whether failed when condition is met."""
        # imitates ansible's Runner class, and should follow the change of the original one.
        register = task_info.vars.get('register')
        if register is not None:
            if 'stdout' in return_data:
                return_data['stdout_lines'] = return_data['stdout'].splitlines()
            task_info.vars[register] = return_data

        failed = False
        reason = None

        # only run the final checks if the async_status has finished,
        # or if we're not running an async_status check at all
        if (task_info.module_name == 'async_status' and "finished" in return_data) \
                or task_info.module_name != 'async_status':
            if failed_when is not None and 'skipped' not in return_data:
                try:
                    failed = check_conditional(failed_when, self.runner.basedir, task_info.vars,
                                               fail_on_undefined=self.runner.error_on_undefined_vars)
                    if failed:
                        reason = 'meet "failed_when" condition'
                except errors.AnsibleError as e:
                    failed = True
                    reason = str(e)

        return failed, reason

    def _show_interpreter(self, task_info, return_data, error_info):
        """ Show an interpreter to debug. """
        next_action = NextAction()

        Interpreter(task_info, return_data, error_info, next_action).cmdloop()
        return next_action




import cmd
import copy
import json
import pprint
import re
import readline
import shlex
import sys

from ansible.callbacks import display
from ansible.playbook.task import Task

from ansibledebugger.interpreter import NextAction
from ansibledebugger import utils


class Interpreter(cmd.Cmd):
    prompt = '(Apdb) '  # Ansible Playbook DeBugger

    def __init__(self, task_info, return_data, error_info, next_action):
        result_decoded = str(error_info.result).decode('unicode-escape')
        self.intro = "The task execution failed.\nreason: %s\nresult: %s\n\nNow a playbook debugger is running..." % (error_info.reason, result_decoded)

        self.task_info = task_info
        self.return_data = return_data
        self.error_info = error_info
        self.next_action = next_action
        cmd.Cmd.__init__(self)

    def cmdloop(self):
        try:
            cmd.Cmd.cmdloop(self)
        except KeyboardInterrupt:
            self.intro = " "
            self.cmdloop()

    do_h = cmd.Cmd.do_help

    def do_error(self, arg):
        """e(rror)
Show an error info.
* 'reason' is the reason the task failed,
* 'result' is the result the module returned. If the module
  throws an expception, 'result' shows the exception.
"""
        result_decoded = str(self.error_info.result).decode('unicode-escape')

        display('reason: %s' % (self.error_info.reason))
        display('result: %s' % (result_decoded))

    do_e = do_error

    def do_list(self, arg):
        """l(ist)
Show the details about this task execution.
* 'module name' is the module name the task executed.
* 'module args' is the simple key=value style args of the module.
* 'complex args' is the module's complex arguments like lists and dicts.
* 'keyword' is the list of keywords the task contains.
  Note that the list may not be complete.
* 'hostname' is the name of a host associated with the task.
  *delegate_to* keyword is not considered here.
* 'groups' is the host's groups.
* 'connection type' is the type of a connection to the host.
* 'ssh host' is the host to ssh. *delegate_to* keyword is
  considered.
* 'ssh options' is the options given to ssh command. Note that
  the options here may not be complete. Run ansible with -vvvv
  option to see the full ssh command.
"""
        template = '{0:<15} : {1}'
        display(template.format('module name', self.task_info.module_name))
        display(template.format('module args', self.task_info.module_args))
        display(template.format('complex args', self.task_info.complex_args))
        display(template.format('keyword', ', '.join(self.get_keyword_list())))

        display(template.format('hostname', self.task_info.vars.get('inventory_hostname', '')))
        groups = self.task_info.vars.get('group_names', [])
        display(template.format('groups', ','.join(groups)))

        connection_type = self.task_info.conn.__module__.split('.')[-1]
        display(template.format('connection type', connection_type))
        if connection_type == 'ssh':
            display(template.format('ssh host', self.task_info.conn.host))
            display(template.format('ssh options', self.task_info.conn.common_args))

    do_l = do_list

    def do_print(self, arg, pp=False):
        """p(rint) [arg]
Print the value of the variable *arg*.

As the special case, if *arg* is `module_name`, print the
module name. Also, if `module_args`, print the simple
key=value style args of the module, and if `complex_args`,
print the complex args like lists and dicts.

With no argument, print all the variables and its value.
"""
        if arg is None or arg == '':
            self.print_module_name(arg, pp)
            self.print_module_args(arg, pp)
            self.print_complex_args(arg, pp)
            self.print_all_vars(pp)
        else:
            if arg == 'module_name':
                self.print_module_name(arg, pp)
            elif arg == 'module_args':
                self.print_module_args(arg, pp)
            elif arg == 'complex_args':
                self.print_complex_args(arg, pp)
            else:
                self.print_var(arg, pp)

    do_p = do_print

    def do_pp(self, arg):
        """pp [arg]
Same as print command, but output is pretty printed.
"""
        self.do_print(arg, pp=True)

    def print_module_name(self, arg, pp=False):
        module_name = self.task_info.module_name
        display('module_name: %s' % (self.pformat_if_pp(module_name, pp)))

    def print_module_args(self, arg, pp=False):
        module_args = self.task_info.module_args
        display('module_args: %s' % (self.pformat_if_pp(module_args, pp)))

    def print_complex_args(self, arg, pp=False):
        complex_args = self.task_info.complex_args
        display('complex_args: %s' % (self.pformat_if_pp(complex_args, pp)))

    def print_var(self, arg, pp=False):
        value = self.task_info.vars.get(arg, 'Not defined')
        display('%s: %s' % (arg, self.pformat_if_pp(value, pp)))

    def print_all_vars(self, pp=False):
        for k, v in self.task_info.vars.iteritems():
            display('%s: %s' % (k, self.pformat_if_pp(v, pp)))

    def pformat_if_pp(self, str, pp):
        if pp:
            return pprint.pformat(str)
        return str

    def do_set(self, arg):
        """set module_args|complex_args key value
Set the argument of the module.

If the first argument is `module_args`, *key*=*value* style
argument is added to the module's args. If the key already
exists, the key is updated. Use quotes if *value* contains
space(s).

If the first argument is `complex_args`, *key* and *value*
are added to module's complex args. *key* specifies the
location where *value* should be added. *key* accepts dot
notation to specify the child of lists and/or dicts.
To update the entire complex_args, use `.` as *key*. *value*
accepts JSON format as well as simple string.
"""
        if arg is None or arg == '':
            display('Invalid option. See help for usage.')
            return

        arg_split = arg.split()
        target = arg_split[0]
        key = arg_split[1]
        index_after_target = arg.find(target) + len(target)
        index_after_key = arg.find(key, index_after_target) + len(key)
        value = arg[index_after_key:].strip()

        if target == 'module_args':
            self.set_module_args(key, value)
        elif target == 'complex_args':
            self.set_complex_args(key, value)
        else:
            display('Invalid option. See help for usage.')

    def set_module_args(self, key, value):
        module_args = self.task_info.module_args
        module_arg_list = shlex.split(module_args)

        replaced = False
        for i, arg in enumerate(module_arg_list):
            if '=' in arg and arg.split('=', 1)[0] == key:
                module_arg_list[i] = '%s=%s' % (key, value)
                replaced = True
                break

        if not replaced:
            module_arg_list.append('%s=%s' % (key, value))

        self.task_info.module_args = ' '.join(module_arg_list)

        display('updated: %s' % (self.task_info.module_args))

    def set_complex_args(self, key, value):
        if key == '.':
            key_list = []
        else:
            key_list = Interpreter.dot_str_to_key_list(key)
            if key_list is None:
                display('Failed to interpret the key')
                return

        try:
            value = json.loads(value)
            value = utils.json_dict_unicode_to_bytes(value)
        except ValueError:
            pass

        if not isinstance(value, dict) and key == '.':
            display('complex_args has to be dict.')
            return

        new_complex_args = copy.deepcopy(self.task_info.complex_args)
        parent = None
        curr = new_complex_args
        last_key = None
        for key, expected_type in key_list:
            if isinstance(curr, dict):
                parent = curr
                last_key = key
                try:
                    curr = curr[key]
                except KeyError:
                    curr = curr[key] = {}

            elif isinstance(curr, list) or isinstance(curr, str):
                try:
                    curr = curr[key]
                except TypeError:
                    curr = parent[last_key] = {}
                    curr = curr[key] = {}
                except IndexError:
                    display('Invalid Index: %s' % str(key))
                    return

                parent = parent[last_key]
                last_key = key

        if parent is None:
            new_complex_args = value
        else:
            parent[last_key] = value
        self.task_info.complex_args = new_complex_args

        display('updated: %s' % (str(self.task_info.complex_args)))

    def do_del(self, arg):
        """del module_args|complex_args key
Delete the argument of the module. The usage of the command's
arguments is almost same as set command.
"""
        if arg is None or arg == '':
            display('Invalid option. See help for usage.')
            return

        arg_split = arg.split()
        target = arg_split[0]
        key = arg_split[1]

        if target == 'module_args':
            self.del_module_args(key)
        elif target == 'complex_args':
            self.del_complex_args(key)
        else:
            display('Invalid option. See help for usage.')

    def del_module_args(self, key):
        module_args = self.task_info.module_args
        module_arg_list = shlex.split(module_args)

        deleted = False
        for i, arg in enumerate(module_arg_list):
            if '=' in arg and arg.split('=', 1)[0] == key:
                del module_arg_list[i]
                display('deleted')
                deleted = True
                break

        if not deleted:
            display('module_args does not contain the key %s' % (key))

        self.task_info.module_args = ' '.join(module_arg_list)

    def del_complex_args(self, key):
        if key == '.':
            key_list = []
        else:
            key_list = Interpreter.dot_str_to_key_list(key)
            if key_list is None:
                display('Failed to interpret the key')
                return

        new_complex_args = copy.deepcopy(self.task_info.complex_args)
        parent = None
        curr = new_complex_args
        last_key = None
        for key, expected_type in key_list:
            parent = curr
            last_key = key
            try:
                curr = curr[key]
            except (KeyError, TypeError, IndexError):
                display('Cannot access the specified element of complex_args. Invalid key: %s' % (key))
                return

        if parent is None:
            new_complex_args = {}
        else:
            del parent[last_key]
        display('deleted')

        self.task_info.complex_args = new_complex_args

    def do_redo(self, args):
        """Re-execute the task, and, if no error, continue to run the playbook.
"""
        self.next_action.set(NextAction.REDO)
        return True

    do_r = do_redo

    def do_EOF(self, line):
        display('')
        return self.do_quit(line)

    def do_quit(self, args):
        """Quit from the debugger. The playbook execution is aborted."""
        self.next_action.set(NextAction.EXIT)
        return True

    do_q = do_quit

    def do_continue(self, arg):
        """Continue without the re-execution of the task."""
        self.next_action.set(NextAction.CONTINUE)
        return True

    do_c = do_cont = do_continue

    def get_keyword_list(self):
        kw_list = []

        if not hasattr(Task, '__slots__'):
            return kw_list

        for slot in Task.__slots__:
            if slot in self.task_info.vars:
                kw = '%s:%s' % (slot, self.task_info.vars[slot])
                kw_list.append(kw)

        return kw_list

    @classmethod
    def dot_str_to_key_list(cls, dot_str):
        """Convert json accessor string in the dot notation to the list of keys.
        Each key in the the returned list includes an expected type.
        """
        pattern = re.compile(r'\A\.?(\w+)|\A\[(\d+)\]')

        key_list = []
        match = pattern.search(dot_str)
        while match is not None:
            if match.group(1) is not None:
                # dict key
                key_list.append((match.group(1), dict))
            elif match.group(2) is not None:
                # list key
                key_list.append((int(match.group(2)), list))
            else:
                break

            dot_str = dot_str[match.end():]
            match = pattern.search(dot_str)

        if dot_str == '':
            return key_list
        else:
            return None





class TaskInfo(object):
    """ Class to describe a target task """
    def __init__(self, conn, tmp_path, module_name, module_args, vars, complex_args):
        self.conn = conn
        self.tmp_path = tmp_path
        self.module_name = module_name
        self.module_args = module_args
        self.vars = vars
        self.complex_args = complex_args


class ErrorInfo(object):
    """ Class to describe the error on a task execution. """
    def __init__(self, failed=False, reason='', result='', error=None):
        self.failed = failed
        self.reason = reason
        self.result = result
        self.error = error


class NextAction(object):
    """ The next action after an interpreter's exit. """
    REDO = 1
    CONTINUE = 2
    EXIT = 3

    def __init__(self, result=EXIT):
        self.result = result

    def set(self, result):
        self.result = result



#!/usr/bin/env python

import os
import sys
sys.path.insert(0, os.path.abspath('lib'))
from setuptools import setup, find_packages

from ansibledebugger import __version__

classifiers = [
    'Environment :: Console',
    'Intended Audience :: Developers',
    'Intended Audience :: System Administrators',
    'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
    'Programming Language :: Python',
    'Topic :: Software Development',
    'Topic :: Software Development :: Debuggers',
    'Topic :: System :: Systems Administration',
]

setup(
    name='ansible-playbook-debugger',
    version=__version__,
    description='Debugger for Ansible Playbook',
    classifiers=classifiers,
    author='Kishin Yagami',
    author_email='k.yagami.suou@gmail.com',
    url='https://github.com/ks888/ansible-playbook-debugger',
    license='GPLv3',
    install_requires=['ansible'],
    packages=find_packages('lib'),
    package_dir={'ansibledebugger': 'lib/ansibledebugger'},
    scripts=[
        'bin/ansible-playbook-debugger',
    ],
    zip_safe=False,
)




import inspect
import os

from ansible import errors
from ansibledebugger.constants import WRAPPED_ACTION_PLUGIN_SUFFIX


def read_module_file(module):
    """ Read the content of the file to which *module* object specifies. """
    src_path = inspect.getsourcefile(module)

    if os.path.exists(src_path):
        with open(src_path) as file:
            module_data = file.read()
    else:
        if hasattr(module_obj, '__loader__') and hasattr(module_obj.__loader__, 'get_data'):
            module_data = module_obj.__loader__.get_data('ansibledebugger/action_plugin_wrapper.py')

        else:
            raise errors.AnsibleError("Failed to load %s module's file" % (module_obj.__name__))

    return module_data


def replace_plugin(plugin, plugin_wrapper, replaced_plugin_dir):
    """ Replace *plugin* module to *plugin_wrapper* module in the
    *replaced_plugin_dir* directory.
    In *replaced_plugin_dir*, the file name of *plugin_wrapper* is that of
    *plugin*. Instead, the original *plugin*'s file name has '_internal' suffix
    so that *plugin_wrapper* is able to call *plugin*.
    """
    plugin_name = plugin.__name__.split('.')[-1]
    plugin_data = read_module_file(plugin)

    wrapped_plugin_name = plugin_name + WRAPPED_ACTION_PLUGIN_SUFFIX + '.py'
    wrapped_plugin_file_path = os.path.join(replaced_plugin_dir,
                                            wrapped_plugin_name)

    with open(wrapped_plugin_file_path, 'w') as file:
        file.write(plugin_data)

    wrapper_name = plugin_name + '.py'
    wrapper_data = read_module_file(plugin_wrapper)
    wrapper_plugin_file_path = os.path.join(replaced_plugin_dir, wrapper_name)

    with open(wrapper_plugin_file_path, 'w') as file:
        file.write(wrapper_data)

# Retrieved from ansible. It's better to use the original one, but its package recentry
# changed, so decided to have another one.
def json_dict_unicode_to_bytes(d):
    ''' Recursively convert dict keys and values to byte str

        Specialized for json return because this only handles, lists, tuples,
        and dict container types (the containers that the json module returns)
    '''

    if isinstance(d, unicode):
        return d.encode('utf-8')
    elif isinstance(d, dict):
        return dict(map(json_dict_unicode_to_bytes, d.iteritems()))
    elif isinstance(d, list):
        return list(map(json_dict_unicode_to_bytes, d))
    elif isinstance(d, tuple):
        return tuple(map(json_dict_unicode_to_bytes, d))
    else:
        return d
